from typing import Optional, Union, overload

import numpy as np
from numpy.typing import NDArray

Number = Union[int, float]


@overload
def nap_to_depth(nap: float, zid: float) -> float:
    ...


@overload
def nap_to_depth(nap: None, zid: float) -> None:
    ...


@overload
def nap_to_depth(
    nap: NDArray[Union[np.floating, np.integer]], zid: float
) -> NDArray[Union[np.floating, np.integer]]:
    ...


def nap_to_depth(
    nap: Optional[Union[Number, NDArray[Union[np.floating, np.integer]]]], zid: float
) -> Optional[Union[float, NDArray[Union[np.floating, np.integer]]]]:
    if nap is None:
        return None
    return -(nap - zid)


@overload
def depth_to_nap(depth: float, zid: float) -> float:
    ...


@overload
def depth_to_nap(depth: None, zid: float) -> None:
    ...


@overload
def depth_to_nap(
    depth: NDArray[Union[np.floating, np.integer]], zid: float
) -> NDArray[Union[np.floating, np.integer]]:
    ...


def depth_to_nap(
    depth: Optional[Union[Number, NDArray[Union[np.floating, np.integer]]]], zid: float
) -> Optional[Union[Number, NDArray[Union[np.floating, np.integer]]]]:
    if depth is None:
        return None
    return zid - depth



from __future__ import annotations

import warnings
from typing import Any, Dict, List, Literal, Mapping, Tuple

import numpy as np
from numpy.typing import NDArray
from pygef.cpt import CPTData
from tqdm import tqdm

transform = {
    "rocks": "G",
    "gravel": "G",
    "sand": "Z",
    "silt": "L",
    "clay": "K",
    "peat": "V",
}


def get_cpt_depth(cpt: CPTData) -> NDArray:
    """
    Returns the array of depth data from the CPT. Uses the (corrected) depth data if
    present, otherwise the `penetrationLength` data along with a warning.
    """
    if "depth" in cpt.data.columns:
        depth = cpt.data["depth"]
    elif "penetrationLength" in cpt.data.columns:
        warnings.warn(
            f"CPT {cpt.alias} has no Corrected Depth trace. py-pilecore uses the Penetration Length instead, which might not be accurate."
        )
        depth = cpt.data["penetrationLength"]
    else:
        raise ValueError(
            f"Cannot process CPT {cpt.alias}: required Corrected Depth or Penetration Length data is missing."
        )

    return np.array(depth, dtype=float)


def create_soil_properties_payload(
    cptdata_objects: List[CPTData],
    classify_tables: Dict[str, dict],
    groundwater_level_nap: float,
    friction_range_strategy: Literal["manual", "lower_bound", "settlement_driven"],
    excavation_depth_nap: float | None = None,
    master_ocr: float | None = None,
    individual_negative_friction_range_nap: Mapping[Any, Tuple[float, float]]
    | None = None,
    individual_positive_friction_range_nap: Mapping[
        Any, Tuple[float, Literal["ptl"] | float]
    ]
    | None = None,
    individual_ocr: Mapping[Any, float] | None = None,
    verbose: bool = False,
) -> Tuple[List[dict], Dict[str, dict]]:
    """
    Creates a dictionary with the `soil_properties` payload content for the PileCore
    endpoints.

    Note
    ------
    the dictionary should be converted to a jsonifyable message before it can be passed
    to a `requests` call directly, for instance with
    `nuclei.client.utils.python_types_to_message()`.

    Parameters
    ----------
    cptdata_objects:
        A list of pygef.CPTData objects
    classify_tables:
        A dictionary, mapping `CPTData.alias` values to dictionary with the resulting response
        of a call to CPTCore `classify/*` information, containing the following keys:
            geotechnicalSoilName: Sequence[str]
                geotechnical Soil Name related to the ISO
            lowerBoundary: Sequence[float]
                lower boundary of the layer [m]
            upperBoundary: Sequence[float]
                upper boundary of the layer [m]
            color: Sequence[str]
                hex color code
            mainComponent: Sequence[Literal["rocks", "gravel", "sand", "silt", "clay", "peat"]]
                main soil component
            cohesion: Sequence[float]
                cohesion of the layer [kPa]
            gamma_sat: Sequence[float]
                Saturated unit weight [kN/m^3]
            gamma_unsat: Sequence[float]
                unsaturated unit weight [kN/m^3]
            phi: Sequence[float]
                phi [degrees]
            undrainedShearStrength: Sequence[float]
                undrained shear strength [kPa]
    groundwater_level_nap:
        The groundwater level. Unit: [m] w.r.t. NAP.
    friction_range_strategy:
        Sets the method to determine the sleeve friction zones on the pile. The soil
        friction in the positive zone contributes to the bearing capacity, while the
        negative zone adds an extra load on the pile. Accepted values: "manual",
        "lower_bound" or "settlement_driven".
    excavation_depth_nap:
        Soil excavation depth after the CPT was taken. Unit: [m] w.r.t. NAP.
    ocr:
        The Over-Consolidation-Ratio [-] of the foundation layer.
    individual_negative_friction_range_nap:
        A dictionary, mapping `CPTData.alias` values to fixed negative-friction ranges.
        For a specification of the values, see `fixed_negative_friction_range_nap`
    individual_positive_friction_range_nap:
        A dictionary, mapping `CPTData.alias` values to fixed positive-friction ranges.
        For a specification of the values, see `fixed_positive_friction_range_nap`
    individual_ocr:
        A dictionary, mapping ``CPTData.alias`` values to Over-Consolidation-Ratio [-]
        values of the foundation layer. This will overrule the general `ocr` setting for
        these specific CPTs only.
    verbose:
        If True, show progress bars and status messages in stdout.

    Returns
    -------
    soil_properties_list:
        The `list_soil_properties` payload content of the PileCore-API multi-cpt
        endpoints.
    results_kwargs:
        Dictionary with keyword arguments for the `pilecore.MultiCPTBearingResults`
        object.
    """
    # Initialize outputs
    results_passover = {}
    soil_properties_list = []

    pbar = None
    if verbose:
        pbar = tqdm(total=len(cptdata_objects))
    for cpt in cptdata_objects:
        # Push verbose message
        if pbar:
            pbar.update()
            pbar.set_description(f"Create soil properties payload for CPT: {cpt.alias}")

        # Construct the cpt_data payload
        cpt_data = dict(
            depth=get_cpt_depth(cpt),
            qc=np.array(cpt.data["coneResistance"], dtype=float).clip(0),
        )

        # Optionally add pore-pressure data to 'cpt_data'.
        if "porePressure" in cpt.data.columns:
            cpt_data["u"] = np.array(cpt.data["porePressure"])

        # Get the layer_table for this cpt from the layer-table dictionary
        if cpt.alias not in classify_tables.keys():
            raise ValueError(f"{cpt.alias} not in `classify_tables`")
        layer_table = classify_tables[cpt.alias]

        # Construct the layer_table_data payload
        layer_table_data = dict(
            depth_btm=layer_table["lowerBoundary"],
            gamma=layer_table["gamma_unsat"],
            gamma_sat=layer_table["gamma_sat"],
            index=list(range(0, len(layer_table["gamma_sat"]))),
            phi=layer_table["phi"],
            soil_code=[transform[soil] for soil in layer_table["mainComponent"]],
            thickness=(
                np.array(layer_table["lowerBoundary"])
                - np.array(layer_table["upperBoundary"])
            ).tolist(),
        )
        # Optionally add consolidation parameters to 'layer_table_data'.
        if "C_p" in layer_table.keys():
            layer_table_data["C_p"] = layer_table["C_p"]
        if "C_s" in layer_table.keys():
            layer_table_data["C_s"] = layer_table["C_s"]

        # Create the Soil-Properties payload
        soil_properties = dict(
            cpt_data=cpt_data,
            layer_table_data=layer_table_data,
            ref_height=cpt.delivered_vertical_position_offset,
            test_id=cpt.alias,
            groundwater_level_nap=groundwater_level_nap
            if groundwater_level_nap is not None
            else cpt.delivered_vertical_position_offset - 1,
        )

        # Optionally add coordinates
        if (
            cpt.delivered_location is not None
            and cpt.delivered_location.x is not None
            and cpt.delivered_location.y is not None
        ):
            soil_properties["coordinates"] = {
                "x": cpt.delivered_location.x,
                "y": cpt.delivered_location.y,
            }

        # Optionally add cpt-specific friction-range parameters
        if (
            individual_negative_friction_range_nap is not None
            and cpt.alias in individual_negative_friction_range_nap.keys()
        ):
            soil_properties[
                "fixed_negative_friction_range_nap"
            ] = individual_negative_friction_range_nap[cpt.alias]
            soil_properties["friction_range_strategy"] = "manual"
        if (
            individual_positive_friction_range_nap is not None
            and cpt.alias in individual_positive_friction_range_nap.keys()
        ):
            soil_properties[
                "fixed_positive_friction_range_nap"
            ] = individual_positive_friction_range_nap[cpt.alias]
            soil_properties["friction_range_strategy"] = "manual"

        # Optionally add OCR parameter
        if individual_ocr is not None and cpt.alias in individual_ocr.keys():
            ocr: float | None = individual_ocr[cpt.alias]
        else:
            ocr = master_ocr
        if ocr is not None:
            soil_properties["ocr"] = ocr

        soil_properties_list.append(soil_properties)
        results_passover[cpt.alias] = {
            "ref_height": cpt.delivered_vertical_position_offset,
            "surface_level_nap": excavation_depth_nap
            if excavation_depth_nap is not None
            else cpt.delivered_vertical_position_offset,
            "location": {"x": cpt.delivered_location.x, "y": cpt.delivered_location.y},
        }

    return soil_properties_list, results_passover



from __future__ import annotations

import datetime
from copy import deepcopy
from typing import Dict, List, Literal, Mapping, Sequence, Tuple

from pygef.cpt import CPTData

from pypilecore.common.piles import PileProperties
from pypilecore.input.soil_properties import create_soil_properties_payload


def create_multi_cpt_payload(
    pile_tip_levels_nap: Sequence[float],
    cptdata_objects: List[CPTData],
    classify_tables: Dict[str, dict],
    groundwater_level_nap: float,
    pile: PileProperties,
    friction_range_strategy: Literal[
        "manual", "lower_bound", "settlement_driven"
    ] = "lower_bound",
    stiff_construction: bool = False,
    cpts_group: List[str] | None = None,
    ocr: float | None = None,
    fixed_negative_friction_range_nap: Tuple[float, float] | None = None,
    fixed_positive_friction_range_nap: Tuple[float, Literal["ptl"] | float]
    | None = None,
    negative_shaft_friction: float | None = None,
    apply_qc3_reduction: bool | None = None,
    relative_pile_load: float | None = 0.7,
    pile_load_sls: float | None = None,
    soil_load_sls: float = 0.0,
    pile_head_level_nap: float | Literal["surface"] = "surface",
    excavation_depth_nap: float | None = None,
    excavation_param_t: float = 1.0,
    excavation_stress_reduction_method: Literal["constant", "begemann"] = "constant",
    excavation_width: float | None = None,
    excavation_edge_distance: float | None = None,
    individual_negative_friction_range_nap: Mapping[str, Tuple[float, float]]
    | None = None,
    individual_positive_friction_range_nap: Mapping[
        str, Tuple[float, Literal["ptl"] | float]
    ]
    | None = None,
    individual_ocr: Mapping[str, float] | None = None,
    use_almere_rules: bool = False,
    overrule_xi: float | None = None,
    gamma_f_nk: float = 1.0,
    gamma_r_s: float = 1.2,
    gamma_r_b: float = 1.2,
) -> Tuple[dict, Dict[str, dict]]:
    """
    Creates a dictionary with the payload content for the PileCore endpoint
    "/compression/multi-cpt/results"

    This dictionary can be passed directly to `nuclei.client.call_endpoint()`.

    Note
    ----
    The result should be converted to a jsonifyable message before it can be passed to a
    `requests` call directly, for instance with
    `nuclei.client.utils.python_types_to_message()`.

    Parameters
    ----------
    pile_tip_levels_nap:
        Range of values for which the pile resistance is calculated.
    cptdata_objects:
        A list of pygef.CPTData objects
    classify_tables:
        A dictionary, mapping `CPTData.alias` values to dictionary with the resulting response
        of a call to CPTCore `classify/*` information, containing the following keys:

            - geotechnicalSoilName: Sequence[str]
                Geotechnical Soil Name related to the ISO
            - lowerBoundary: Sequence[float]
                Lower boundary of the layer [m]
            - upperBoundary: Sequence[float]
                Upper boundary of the layer [m]
            - color: Sequence[str]
                Hex color code
            - mainComponent: Sequence[Literal["rocks", "gravel", "sand", "silt", "clay", "peat"]]
                Main soil component
            - cohesion: Sequence[float]
                Cohesion of the layer [kPa]
            - gamma_sat: Sequence[float]
                Saturated unit weight [kN/m^3]
            - gamma_unsat: Sequence[float]
                Unsaturated unit weight [kN/m^3]
            - phi: Sequence[float]
                Phi [degrees]
            - undrainedShearStrength: Sequence[float]
                Undrained shear strength [kPa]

    groundwater_level_nap:
        The groundwater level. Unit: [m] w.r.t. NAP.
    pile:
        A PileProperties object.
    friction_range_strategy:
        Sets the method to determine the sleeve friction zones on the pile. The soil
        friction in the positive zone contributes to the bearing capacity, while the
        negative zone adds an extra load on the pile. Accepted values: "manual",
        "lower_bound" (default) or "settlement_driven".
    stiff_construction:
        Set to True if it's a stiff costruction. This will have influence on the xi factor
        if you don't overrule it. Default = False.
    cpts_group:
        CPTs that are considered one group. Items must relate to the alias of the CPTData
        objects in `cptdata_objects`.
    ocr:
        The Over-Consolidation-Ratio [-] of the foundation layer.
    fixed_negative_friction_range_nap:
        Optionally sets the fixed depth range between which the negative sleeve friction
        is calculated. If an array of format [top, bottom], the range is set between top
        and bottom where top and bottom are floating values.
        Unit: [m] w.r.t. NAP
    fixed_positive_friction_range_nap:
        Optionally sets the fixed depth range between which the positive sleeve friction
        is calculated. If an array of format (top, bottom), the range is set between top
        and bottom where top and bottom are floating point values. If bottom == "ptl",
        the pile tip level of the calculation is used as value for bottom.
        Unit: [m] w.r.t. NAP
    negative_shaft_friction:
        Sets a fixed value for the negative friction force. If provided, the
        fixed_negative_friction_range_nap parameter will be ignored.
        Unit: [kN]
    apply_qc3_reduction:
        Determines if the reduction on the qc;III trajectory for auger piles should be
        applied conform 7.6.2.3(e) of NEN 9997-1. If omitted (or null: default), the
        value is inferred from the pile type: only True for auger piles (when
        pile_properties.is_auger == True). If a boolean is provided, the qc3 reduction
        is/isn't applied, regardless of the pile type.
    relative_pile_load:
        The fraction of the maximum bearing capacity that is used as pilehead force in
        the settlement calculations. The input value can range between 0.0 and 1.0,
        where 1.0 translates to a pile load of 100% of the maximum bearing capacity.
        When the pile_load_sls parameter is provided, rel_pile_load will be ignored.

        When multiple pile-tip levels are considered, the applied pile load will vary
        with each pile-tip level, depending on the bearing capacity.

        Default = 0.7
    pile_load_sls:
        Force on pile in SLS [kN]. Used to determine settlement of pile w.r.t. soil.
    soil_load_sls:
        Load on soil surface [kPa], used to calculate soil settlement. This is only required
        with the settlement-driven friction-range strategy.
        Default = 0.0
    pile_head_level_nap:
        The level of the pile head. Can be:

            - float;
                This is interpreted as an absolute level in [m w.r.t. NAP];
            - Literal["surface"] (default).
                In this case, the soil_properties.service_level
                property is used.
    excavation_depth_nap:
        Soil excavation depth after the CPT was taken. Unit: [m] w.r.t. NAP.
    excavation_param_t:
        Excavation parameter depending on pile installation method and/or phasing.
            - Use 1.0 (default) for post-excavation installation with vibration (i.e. hammering).
            - Use 0.5 for reduced-vibration installation, or pile installation prior to
              excavation method.

        See for more info NEN 9997-1+C2:2017 7.6.2.3.(10)(k)
    excavation_stress_reduction_method:
        Method used to calculate the stress reduction due to the excavation applied to the effective and total stresses.
        Only used when `excavation_depth_nap` is different than `None`. It can be:
            - "constant" (default): The stress reduction below the excavation is constant with depth. The stress reduction
            is equal to the original effective stress (i.e. before the excavation) at the excavation depth.
            - "begemann": The stress reduction below the excavation decreases with depth according to the Begemann method.
            This method uses the elastic solution of a strip load acting on a semi-infinite homogeneous soil mass.
            The load has a width equal to the excavation width and a magnitude equal to the original effective stress at
            the excavation depth.
        Regardless the method, the stress reduction applied above the excavation is equal to the original effective stress
        at each corresponding depth.
    excavation_width:
        Width of the excavation [m]. Used to calculate the stress reduction due to the excavation if the Begemann method is selected.
        In this case, it must be provided and it must be > 0.
    excavation_edge_distance:
        Distance from the pile centerline to the excavation edge [m]. Used to calculate the stress reduction due to the excavation if
        the Begemann method is selected. In this case, it must be provided and it must be between 0.0 and 0.5 * excavation_width.

        Note that:
            - 0.0 means that the pile is located at the edge of the excavation.
            - 0.5 * excavation_width means that the pile is at the center of the excavation.
    individual_negative_friction_range_nap:
        A dictionary, mapping ``CPTData.alias`` values to fixed negative-friction ranges.
        For a specification of the values, see ``fixed_negative_friction_range_nap``
    individual_positive_friction_range_nap:
        A dictionary, mapping ``CPTData.alias`` values to fixed positive-friction ranges.
        For a specification of the values, see ``fixed_positive_friction_range_nap``
    individual_ocr:
        A dictionary, mapping ``CPTData.alias`` values to Over-Consolidation-Ratio [-]
        values of the foundation layer. This will overrule the general `ocr` setting for
        these specific CPTs only.
    use_almere_rules:
        If set to True the contribution, produced by the positive shaft friction, to the
        total bearing capacity is limited to at most 75% the contribution provided by
        the pile tip. ref: https://www.almere.nl/fileadmin/user_upload/Richtlijnen_Constructie_Gem._Almere_vanaf_01-01-2017_versie_3.0a.pdf
    overrule_xi:
        Set a fixed value for xi in all calculations. Use with caution. This will
        overrule the calculation of xi-values based on the group-size, variation-
        coefficient and construction stiffness.
    gamma_f_nk:
        Safety factor for design-values of the negative sleeve friction force.
        Default = 1.0
    gamma_r_s:
        Safety factor, used to obtain design-values of the pile-tip bearingcapacity.
        Default = 1.2
    gamma_r_b:
        Safety factor, used to obtain design-values of the sleeve bearingcapacity.
        Default = 1.2

    Returns
    -------
    multi_cpt_payload:
        Dictionary with the payload content for the PileCore endpoint
        "/compression/multi-cpt/results"
    results_kwargs:
        Dictionary with keyword arguments for the `pilecore.MultiCPTBearingResults`
        object.

    Raises
    ------
    ValueError:
        - If `excavation_depth_nap` is not None and `excavation_param_t` is None.
        - If `excavation_stress_reduction_method` is not either 'constant' or 'begemann'.
        - If `excavation_stress_reduction_method` is 'begemann' and `excavation_width` is None.
        - If `excavation_stress_reduction_method` is 'begemann' and `excavation_edge_distance` is None.
        - If both `relative_pile_load` and `pile_load_sls` are None.
    """
    # Input validation
    if excavation_depth_nap is not None and excavation_param_t is None:
        raise ValueError(
            "`excavation_param_t` cannot be None when `excavation_depth_nap` is not None."
        )
    if excavation_stress_reduction_method not in ["constant", "begemann"]:
        raise ValueError(
            "`excavation_stress_reduction_method` must be either 'constant' or 'begemann'."
        )
    if excavation_stress_reduction_method == "begemann" and excavation_width is None:
        raise ValueError(
            "`excavation_width` must be provided when `excavation_stress_reduction_method` is 'begemann'."
        )
    if (
        excavation_stress_reduction_method == "begemann"
        and excavation_edge_distance is None
    ):
        raise ValueError(
            "`excavation_edge_distance` must be provided when `excavation_stress_reduction_method` is 'begemann'"
        )
    if relative_pile_load is None and pile_load_sls is None:
        raise ValueError(
            "Need at least a value for one of: [`relative_pile_load`, `pile_load_sls`]"
        )

    soil_properties_list, results_kwargs = create_soil_properties_payload(
        cptdata_objects=cptdata_objects,
        classify_tables=classify_tables,
        groundwater_level_nap=groundwater_level_nap,
        friction_range_strategy=friction_range_strategy,
        excavation_depth_nap=excavation_depth_nap,
        master_ocr=ocr,
        individual_negative_friction_range_nap=individual_negative_friction_range_nap,
        individual_positive_friction_range_nap=individual_positive_friction_range_nap,
        individual_ocr=individual_ocr,
    )
    multi_cpt_payload: dict = dict(
        pile_tip_levels_nap=list(pile_tip_levels_nap),
        list_soil_properties=soil_properties_list,
        pile_properties=pile.serialize_payload(),
        friction_range_strategy=friction_range_strategy,
        pile_head_level_nap=pile_head_level_nap,
        stiff_construction=stiff_construction,
        rel_pile_load=relative_pile_load,
        soil_load=soil_load_sls if soil_load_sls is not None else 0.0,
        excavation_param_t=excavation_param_t,
        use_almere_rules=use_almere_rules,
        gamma_f_nk=gamma_f_nk,
        gamma_r_b=gamma_r_b,
        gamma_r_s=gamma_r_s,
    )

    # Add optional properties
    if excavation_depth_nap is not None:
        multi_cpt_payload["excavation_depth_nap"] = excavation_depth_nap

    if excavation_stress_reduction_method == "constant":
        multi_cpt_payload["excavation_settings"] = dict(
            stress_reduction_method="constant"
        )
    else:
        multi_cpt_payload["excavation_settings"] = dict(
            stress_reductin_method="begemann",
            excavation_width=excavation_width,
            excavation_edge_distance=excavation_edge_distance,
        )

    if pile_load_sls is not None:
        multi_cpt_payload["pile_load"] = pile_load_sls
    if apply_qc3_reduction is not None:
        multi_cpt_payload["apply_qc3_reduction"] = apply_qc3_reduction
    if negative_shaft_friction is not None:
        multi_cpt_payload["f_nk"] = negative_shaft_friction
    if fixed_negative_friction_range_nap is not None:
        multi_cpt_payload[
            "fixed_negative_friction_range_nap"
        ] = fixed_negative_friction_range_nap
    if fixed_positive_friction_range_nap is not None:
        multi_cpt_payload[
            "fixed_positive_friction_range_nap"
        ] = fixed_positive_friction_range_nap
    if overrule_xi is not None:
        multi_cpt_payload["overrule_xi"] = overrule_xi
    if cpts_group is not None:
        multi_cpt_payload["cpts_group"] = cpts_group

    return multi_cpt_payload, results_kwargs


def create_multi_cpt_report_payload(
    multi_cpt_payload: dict,
    project_name: str,
    project_id: str,
    author: str,
    date: str | None = None,
    group_results_content: bool = True,
    individual_cpt_results_content: bool = True,
    result_summary_content: bool = True,
) -> dict:
    """
    Creates a dictionary with the payload content for the PileCore endpoint
    "/compression/multi-cpt/report"

    This dictionary can be passed directly to `nuclei.client.call_endpoint()`. Note that
    it should be converted to a jsonifyable message before it can be passed to a
    `requests` call directly, for instance with
    `nuclei.client.utils.python_types_to_message()`.

    Parameters
    ----------
    multi_cpt_payload:
        Index 0 of the resulting tuple of a call to `create_multi_cpt_payload()`
    project_name:
        The name of the project.
    project_id:
        The identifier (code) of the project.
    author:
        The author of the report.
    date:
        The date of the report. If None, the current date will be used.
    group_results_content:
        Whether or not to add the group-results section to the report. Default = True
    individual_cpt_results_content:
        Whether or not to add the individual-cpt-results section to the report.
        Default = True
    result_summary_content:
        Whether or not to add the result-summary section to the report. Default = True

    Returns
    -------
    report_payload:
        Dictionary with the payload content for the PileCore endpoint
        "/compression/multi-cpt/report"
    """
    report_payload = deepcopy(multi_cpt_payload)
    report_payload.update(
        dict(
            content=dict(
                group_results=group_results_content,
                individual_cpt_results=individual_cpt_results_content,
                result_summary=result_summary_content,
            ),
            general=dict(
                author=author,
                project_id=project_id,
                project_name=project_name,
                date=date
                if date is None
                else datetime.date.today().strftime("%d-%m-%y"),
            ),
        )
    )
    return report_payload



from __future__ import annotations

from typing import Dict, List

from pypilecore.common.piles.geometry import PileGeometry
from pypilecore.common.piles.geometry.components import (
    PrimaryPileComponentDimension,
    RectPileGeometryComponent,
    RoundPileGeometryComponent,
)
from pypilecore.common.piles.geometry.materials import MATERIALS, PileMaterial
from pypilecore.common.piles.type import PileType


class PileProperties:
    """The PileProperties class represents all properties of a pile."""

    def __init__(
        self, geometry: PileGeometry, pile_type: PileType, name: str | None = None
    ):
        """
        Represents all properties of a pile.

        Parameters:
        -----------
        geometry: PileGeometry
            The geometry of the pile.
        pile_type: PileType
            The type of the pile.
        name: str, optional
            The name of the pile, by default None.
        """
        self._geometry = geometry
        self._pile_type = pile_type
        self._name = name

    @classmethod
    def from_api_response(cls, pile_properties: dict) -> PileProperties:
        """
        Instantiates a PileProperties from a response dictionary.

        Parameters:
        -----------
        pile_properties: dict
            A dictionary that is retrieved from the API response payload at "/pile_properties".

        Returns:
        --------
        PileProperties
            A pile properties object.
        """
        return cls(
            geometry=PileGeometry.from_api_response(pile_properties["geometry"]),
            pile_type=PileType.from_api_response(pile_properties["pile_type"]),
            name=pile_properties.get("name"),
        )

    @property
    def geometry(self) -> PileGeometry:
        """The geometry of the pile."""
        return self._geometry

    @property
    def pile_type(self) -> PileType:
        """The type of the pile."""
        return self._pile_type

    @property
    def name(self) -> str | None:
        """The name of the pile."""
        return self._name

    def serialize_payload(self) -> Dict[str, dict | str]:
        """
        Serialize the pile properties to a dictionary payload for the API.

        Returns:
            A dictionary payload containing the geometry, pile type, and name (if set).
        """
        payload: Dict[str, dict | str] = {
            "geometry": self.geometry.serialize_payload(),
            "pile_type": self.pile_type.serialize_payload(),
        }

        if self.name is not None:
            payload["name"] = self.name

        return payload


def create_basic_pile(
    pile_shape: str,
    pile_name: str | None = None,
    main_type: str | None = None,
    specification: str | int | float | None = None,
    installation: str | None = None,
    height_base: float | None = None,
    core_secondary_dimension: float | None = None,
    core_tertiary_dimension: float | None = None,
    base_secondary_dimension: float | None = None,
    base_tertiary_dimension: float | None = None,
    core_diameter: float | None = None,
    base_diameter: float | None = None,
    pile_material: str | None = None,
    custom_material: dict | None = None,
    settlement_curve: int | None = None,
    adhesion: float | None = None,
    alpha_p: float | None = None,
    alpha_s_clay: float | None = None,
    alpha_s_sand: float | None = None,
    beta_p: float | None = None,
    pile_tip_factor_s: float | None = None,
    is_auger: bool | None = None,
    is_low_vibrating: bool | None = None,
    negative_fr_delta_factor: float | None = None,
) -> PileProperties:
    """
    Create a basic pile with the provided parameters.
    A "basic" pile is a pile with a basic geometry, containing at most 2 components,
    namely a core component and an optional shell/widened-base component.

    Parameters
    ----------
    pile_shape : str
        The shape of the pile. Either "round" or "rect".
    pile_name : str, optional
        The name of the pile, by default None.
    main_type : str, optional
        The main type of the standard pile type definition, by default None.
    specification : str | int | float, optional
        The specification of the standard pile type definition, by default None.
        Required if `main_type` is provided.
    installation : str, optional
        The installation of the standard pile type definition, by default None.
    height_base : float, optional
        The height [m] of the base component, by default None.
        Required if base dimensions are provided.
    core_secondary_dimension : float, optional
        The largest dimension [m] of the core component's cross-section, by default None.
        Required for rectangular piles.
    core_tertiary_dimension : float, optional
        The smallest dimension [m] of the core component's cross-section, by default None.
    base_secondary_dimension : float, optional
        The largest dimension [m] of the base/shell component's cross-section, by default None.
    base_tertiary_dimension : float, optional
        The smallest dimension [m] of the base/shell component's cross-section, by default None.
    core_diameter : float, optional
        The diameter [m] of the core component's cross-section, by default None.
        Required for round piles.
    base_diameter : float, optional
        The diameter [m] of the base/shell component's cross-section, by default None.
    pile_material : str, optional
        The material of the pile, by default None.
        Required if no standard pile type was specified.
    custom_material : dict, optional
        A custom material for the pile, by default None.
    settlement_curve : int, optional
        The settlement curve of the pile, by default None.
        Required if no standard pile type was specified.
    adhesion : float, optional
        The adhesion of the pile, by default None.
        Required if no standard pile type was specified.
    alpha_p : float, optional
        The alpha_p of the pile, by default None.
        Required if no standard pile type was specified.
    alpha_s_clay : float, optional
        The alpha_s_clay of the pile, by default None.
        Required if no standard pile type was specified.
    alpha_s_sand : float, optional
        The alpha_s_sand of the pile, by default None.
        Required if no standard pile type was specified.
    beta_p : float, optional
        The beta_p of the pile, by default None.
        Required if no standard pile type was specified.
    pile_tip_factor_s : float, optional
        The pile tip factor s of the pile, by default None.
        Required if no standard pile type was specified.
    is_auger : bool, optional
        Whether the pile is an auger pile, by default None.
        Required if no standard pile type was specified.
    is_low_vibrating : bool, optional
        Whether the pile is low vibrating, by default None.
        Required if no standard pile type was specified.
    negative_fr_delta_factor : float, optional
        The negative friction delta factor of the pile, by default None.
        Required if no standard pile type was specified.

    Returns
    -------
    PileProperties
        The pile properties object.
    """
    materials = [
        PileMaterial(**material_value) for material_value in MATERIALS.values()  # type: ignore
    ]

    if custom_material is not None:
        materials = [PileMaterial(**custom_material)]

    if pile_shape == "round":
        if core_diameter is None:
            raise ValueError("`core_diameter` must be provided for a round pile.")

        components: List[RoundPileGeometryComponent | RectPileGeometryComponent] = [
            RoundPileGeometryComponent(
                diameter=core_diameter,
                primary_dimension=PrimaryPileComponentDimension(
                    length=None,
                ),
                material=pile_material,
            )
        ]

        if base_diameter is not None:
            components.append(
                RoundPileGeometryComponent(
                    diameter=base_diameter,
                    primary_dimension=PrimaryPileComponentDimension(
                        length=height_base,
                    ),
                    material=pile_material,
                )
            )

    elif pile_shape in ["rect", "rectangle"]:
        if core_secondary_dimension is None:
            raise ValueError(
                "`core_secondary_dimension` must be provided for a rectangular pile."
            )

        components = [
            RectPileGeometryComponent(
                primary_dimension=PrimaryPileComponentDimension(length=None),
                secondary_dimension=core_secondary_dimension,
                tertiary_dimension=core_tertiary_dimension,
                material=pile_material,
            )
        ]

        if base_secondary_dimension is not None:
            components.append(
                RectPileGeometryComponent(
                    secondary_dimension=base_secondary_dimension,
                    tertiary_dimension=base_tertiary_dimension,
                    primary_dimension=PrimaryPileComponentDimension(
                        length=height_base,
                    ),
                    material=pile_material,
                )
            )

    else:
        raise ValueError(f"Invalid pile shape: {pile_shape}")

    geometry = PileGeometry(
        components=components,
        materials=materials,
        pile_tip_factor_s=pile_tip_factor_s,
        beta_p=beta_p,
    )

    if main_type is not None:
        if specification is None:
            raise ValueError("Specification must be provided if main type is provided.")

        standard_pile: Dict[str, str | int] | None = {
            "main_type": main_type,
            "specification": int(specification),
        }

        if installation is not None:
            standard_pile["installation"] = installation  # type: ignore

    else:
        standard_pile = None

    pile_type = PileType(
        standard_pile=standard_pile,
        alpha_s_sand=alpha_s_sand,
        alpha_s_clay=alpha_s_clay,
        alpha_p=alpha_p,
        alpha_t_sand=None,
        settlement_curve=settlement_curve,
        negative_fr_delta_factor=negative_fr_delta_factor,
        adhesion=adhesion,
        is_low_vibrating=is_low_vibrating,
        is_auger=is_auger,
    )

    return PileProperties(geometry=geometry, pile_type=pile_type, name=pile_name)



from __future__ import annotations  # noqa: F404

from typing import Hashable, List

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from natsort import natsort_keygen

from pypilecore.results.cases_multi_cpt_results import CasesMultiCPTBearingResults
from pypilecore.results.result_definitions import CPTGroupResultDefinitions


class FigureCPTGroupResultsVersusPtls:
    """
    Interactive figure to show the CPT grouped results of the bearing capacity calculations
    versus the pile tip levels (PTLs).

    The layout of the figure is:
        - X axis: result values.
        - Y axis: pile tip level w.r.t. NAP.
        - Each trace represents a different case.

    The figure has a method to switch between results.
    """

    def __init__(self, cases_multi_results: CasesMultiCPTBearingResults) -> None:
        """
        Initializes the figure.

        Parameters
        ----------
        cases_multi_results : CasesMultiCPTBearingResults
            The results of the bearing capacity calculations.

        Raises
        ------
        TypeError
            If `cases_multi_results` are not of type 'CasesMultiCPTBearingResults'.
        """
        # Validate the data
        self._set_results(cases_multi_results)

        # Initialize the figure
        self._figure = go.FigureWidget(
            layout=go.Layout(
                height=800,
                width=800,
                title="CPT Group Results vs. Pile tip level for all cases<br>Result: ",
                legend_title="Case",
                colorway=px.colors.qualitative.Plotly,
                xaxis_title="",
                yaxis=go.layout.YAxis(
                    title="Pile tip level [m NAP]",
                    title_font_size=18,
                ),
                autosize=False,
            )
        )

    def _set_results(self, value: CasesMultiCPTBearingResults) -> None:
        """Private setter for the results."""
        if not isinstance(value, CasesMultiCPTBearingResults):
            raise TypeError(
                f"Expected type 'CasesMultiCPTBearingResults' for 'cases_multi_results', but got {type(value)}"
            )
        self._results = value

    @property
    def results(self) -> CasesMultiCPTBearingResults:
        """The results of the bearing capacity calculations."""
        return self._results

    @property
    def data(self) -> pd.DataFrame:
        """The dataframe used to plot the results."""
        return self.results.cpt_group_results_dataframe

    @property
    def cases(self) -> List[Hashable]:
        """The case names of each MultiCPTBearingResults."""
        return self.results.cases

    @property
    def figure(self) -> go.FigureWidget:
        """The figure widget."""
        return self._figure

    def get_visible_cases(self) -> List[Hashable]:
        """
        Returns the visible cases in the figure widget.

        Notes
        -----
            - If there are no traces in the figure, it returns all the case names.
            - If a case is None, then it is represented as "None".
        """
        if len(self.figure.data) == 0:
            # Select all cases if there are no traces in the figure yet.
            cases: List[Hashable] = []
            for case in self.cases:
                if case is None:
                    cases.append("None")
                else:
                    cases.append(case)
            return cases
        return [trace.name for trace in self.figure.data if trace.visible is True]

    def show_result(self, result_name: str) -> None:
        """Shows the group results for all pile tip levels for the requested `result_name`.

        Parameters
        ----------
        result_name : str
            The name of the result to show.

        Raises
        ------
        ValueError
            If the `result_name` is not found in the CPTResultDefinitions.
        """
        # Get the result definition that corresponds to the result name.
        result_definition = CPTGroupResultDefinitions.get(result_name)

        # Get the visible cases
        visible_cases = self.get_visible_cases()

        # Select data for result type
        selected_data = self.data.loc[
            self.data["result_name"] == result_definition.name
        ]

        # Organize data and format depending on result type
        mode = "lines+markers"
        marker_size = 6
        if result_definition in [
            CPTGroupResultDefinitions.cpt_Rc_min,
            CPTGroupResultDefinitions.cpt_Rc_max,
            CPTGroupResultDefinitions.cpt_normative,
            CPTGroupResultDefinitions.use_group_average,
            CPTGroupResultDefinitions.xi_normative,
            CPTGroupResultDefinitions.n_cpts,
        ]:
            selected_data = selected_data.sort_values(by="result", key=natsort_keygen())
            mode = "markers"
            marker_size = 10

        traces = []
        for case in self.cases:
            if case is not None:
                df = selected_data.loc[selected_data["case_name"] == case]
            else:
                df = selected_data.loc[selected_data["case_name"].isna()]

            traces.append(
                go.Scatter(
                    x=df["result"],
                    y=df["pile_tip_level_nap"],
                    mode=mode,
                    name=case if case is not None else "None",
                    marker_size=marker_size,
                )
            )

        with self.figure.batch_update():
            # Empty traces
            self.figure.data = []

            # Apply changes
            self.figure.add_traces(traces)

            self.figure.update_layout(
                title=f"CPT Group Results vs. Pile tip level for all cases<br>Result: {result_definition.value.html}",
                xaxis=go.layout.XAxis(
                    title=f"{result_definition.value.html} [{result_definition.value.unit}]",
                    title_font_size=18,
                ),
                showlegend=True,
            )

            self.figure.update_traces(
                selector=lambda x: x.name in visible_cases,
                patch=dict(
                    visible=True,
                ),
            )

            self.figure.update_traces(
                selector=lambda x: x.name not in visible_cases,
                patch=dict(
                    visible="legendonly",
                ),
            )



from pypilecore.input.grouper_properties import (
    create_grouper_payload,
    create_grouper_report_payload,
)
from pypilecore.input.multi_cpt import (
    create_multi_cpt_payload,
    create_multi_cpt_report_payload,
)
from pypilecore.input.soil_properties import create_soil_properties_payload

__all__ = [
    "create_soil_properties_payload",
    "create_multi_cpt_payload",
    "create_grouper_payload",
    "create_grouper_report_payload",
    "create_multi_cpt_report_payload",
]



from pypilecore.results.cases_multi_cpt_results import CasesMultiCPTBearingResults
from pypilecore.results.grouper_result import GrouperResults
from pypilecore.results.multi_cpt_results import MultiCPTBearingResults
from pypilecore.results.single_cpt_results import SingleCPTBearingResults
from pypilecore.results.soil_properties import SoilProperties

__all__ = [
    "CasesMultiCPTBearingResults",
    "MultiCPTBearingResults",
    "SingleCPTBearingResults",
    "SoilProperties",
    "GrouperResults",
]



from __future__ import annotations

from copy import deepcopy
from typing import Any, Tuple

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from numpy.typing import NDArray

# Axis with mobilized capacity as a fracion of maximum.
mob_over_max = np.array(
    [
        0,
        0.10,
        0.15,
        0.20,
        0.25,
        0.30,
        0.35,
        0.40,
        0.45,
        0.50,
        0.55,
        0.60,
        0.65,
        0.70,
        0.75,
        0.80,
        0.85,
        0.90,
        0.95,
        0.98,
        1.00,
    ]
)

# Axis with s_b/Deq [%] due to pile-tip mobilization for settlement-curve 1
sb_rb_1 = np.array(
    [
        0,
        0.03,
        0.08,
        0.16,
        0.26,
        0.4,
        0.55,
        0.73,
        0.95,
        1.18,
        1.48,
        1.83,
        2.3,
        2.85,
        3.49,
        4.25,
        5.13,
        6.33,
        7.8,
        9,
        10,
    ]
)

# Axis with s_b/Deq [%] due to pile-tip mobilization for settlement-curve 2
sb_rb_2 = np.array(
    [
        0,
        0.15,
        0.3,
        0.5,
        0.75,
        1.06,
        1.41,
        1.79,
        2.2,
        2.71,
        3.29,
        3.99,
        4.81,
        5.8,
        7.04,
        8.57,
        10.46,
        12.76,
        15.62,
        17.95,
        20,
    ]
)

# Axis with s_b/Deq [%] due to pile-tip mobilization for settlement-curve 3
sb_rb_3 = np.array(
    [
        0,
        0.55,
        0.91,
        1.31,
        1.73,
        2.19,
        2.7,
        3.23,
        3.78,
        4.4,
        5.09,
        5.84,
        6.85,
        8.07,
        9.45,
        10.99,
        12.73,
        14.67,
        17.05,
        18.7,
        20,
    ]
)

# Axis with s_b [mm] due to pile-shaft mobilization for settlement-curve 1
sb_rs_1 = np.array(
    [
        0,
        0.05,
        0.1,
        0.17,
        0.29,
        0.45,
        0.65,
        0.87,
        1.12,
        1.39,
        1.69,
        2.02,
        2.38,
        2.78,
        3.3,
        4.04,
        4.96,
        6.22,
        7.87,
        9.16,
        10.3,
    ]
)

# Axis with s_b [mm] due to pile-shaft mobilization for settlement-curve 2
sb_rs_2 = np.array(
    [
        0,
        0.55,
        0.91,
        1.3,
        1.73,
        2.18,
        2.67,
        3.24,
        3.82,
        4.47,
        5.16,
        5.91,
        6.7,
        7.51,
        8.59,
        9.97,
        11.83,
        14.48,
        17.93,
        20.8,
        23.5,
    ]
)

# Axis with s_b [mm] due to pile-shaft mobilization for settlement-curve 3
sb_rs_3 = np.array(
    [
        0,
        0.55,
        0.91,
        1.3,
        1.73,
        2.18,
        2.67,
        3.24,
        3.82,
        4.47,
        5.16,
        5.91,
        6.7,
        7.51,
        8.59,
        9.97,
        11.83,
        14.48,
        17.93,
        20.8,
        23.5,
    ]
)

# Axis with s_b [m] due to pile-shaft mobilization for settlement-curve 1 [ANCHOR]
mob_over_max_anchor = np.array(
    [
        0,
        0.25,
        0.4,
        0.5,
        0.75,
        1.0,
        1.5,
        2.0,
        3.0,
        4.0,
        5.0,
        6.0,
        7.0,
        7.5,
        8.0,
        9.0,
        9.5,
        10,
        10.6,
        11.0,
        12.0,
        13.0,
        14.0,
        15.0,
        16.0,
        17.0,
        18.0,
        19.0,
        20,
        21.0,
        22.0,
        23.0,
        24.0,
        25.0,
        26.0,
        27.0,
        28.0,
        29.0,
        30.0,
        50.0,
    ]
)

# Axis with mobilized capacity as a fraction of maximum for settlement-curve 1 [ANCHOR]
sb_rs_1_anchor = np.array(
    [
        0.0,
        0.25,
        0.29,
        0.32,
        0.40,
        0.44,
        0.53,
        0.60,
        0.73,
        0.81,
        0.865,
        0.902,
        0.930,
        0.942,
        0.954,
        0.976,
        0.987,
        1.0,
        1.0,
        1.0,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
    ]
)

# Axis with mobilized capacity as a fraction of maximum for settlement-curve 2 [ANCHOR]
sb_rs_2_anchor = np.array(
    [
        0.0,
        0.048,
        0.06,
        0.08,
        0.11,
        0.16,
        0.23,
        0.29,
        0.39,
        0.47,
        0.55,
        0.61,
        0.668,
        0.69,
        0.723,
        0.77,
        0.79,
        0.805,
        0.825,
        0.835,
        0.863,
        0.885,
        0.902,
        0.919,
        0.93,
        0.943,
        0.954,
        0.964,
        0.972,
        0.98,
        0.984,
        0.99,
        0.995,
        1.0,
        1.0,
        0.50,
        0.50,
        0.50,
        0.50,
        0.50,
    ]
)

curve_type_map = {
    1: {
        "sb_rs": sb_rs_1,
        "sb_rb": sb_rb_1,
    },
    2: {
        "sb_rs": sb_rs_2,
        "sb_rb": sb_rb_2,
    },
    3: {
        "sb_rs": sb_rs_3,
        "sb_rb": sb_rb_3,
    },
}


def get_load_settlement_axes_data(
    curve_type: int,
) -> Tuple[NDArray[np.floating], NDArray[np.floating], NDArray[np.floating]]:
    """
    Returns the axis values from the settlement lines in Figure 7.n & 7.o

    Parameters
    ----------
    curve_type:
        The settlement-curve type

    Returns
    -------
    mob_ratio
        Array of mobilized bearing capacity ratios
    sb_rb
        Array of settlements due to Rb-mobilization, as a percentage of Deq.
    sb_rs
        Array of settlements due to Rs-mobilization [mm].
    """

    mob_ratio = deepcopy(mob_over_max)

    curves = curve_type_map[curve_type]

    return mob_ratio, curves["sb_rb"], curves["sb_rs"]


def get_load_settlement_plot(
    settlement_curve: int,
    d_eq: float,
    s_b: float,
    f_c_k: float,
    f_nk_k: float,
    r_b_k: float,
    r_s_k: float,
    rs_mob_ratio: float,
    rb_mob_ratio: float,
    sb_max: float = 25,
    axes: Axes | None = None,
    figsize: Tuple[float, float] = (6.0, 6.0),
    **kwargs: Any,
) -> Axes:
    """
    Returns a `Axes` object with a load-settlement diagram. Based on NEN 9997-1+C2
    section 7.6.4.2(i).

    Parameters
    ----------
    settlement_curve: int
        The settlement curve of the pile, according to table 7.c of NEN 9997-1+C2.
    d_eq: float
        The equivalent diameter of the pile base [m].
    s_b: float
        The settlement at pile-tip level [mm].
    f_c_k: float
        The compressive force on the pile-head [kN].
    f_nk_k: float
        The characteristic negative friction force [kN].
    r_b_k: float
        The characteristic pile-tip resistance [kN].
    r_s_k: float
        The characteristic shaft resistance [kN].
    rs_mob_ratio: float
        Ratio of the mobilized shaft resistance [-].
    rb_mob_ratio: float
        Ratio of the mobilized base resistance [-].
    sb_max: float
        Maximum pile-tip settlement [mm].
    axes:
        Optional `Axes` object. If not provided, a `plt.Figure`
        and `Axes` objects will be created.
    figsize: Tuple[float, float]
        The figure dimensions. Default = (6.0, 6.0)
    **kwargs:
        All additional keyword arguments are passed to the `pyplot.figure` call.

    """
    # Create axes objects if not provided
    if axes is not None:
        if not isinstance(axes, Axes):
            raise ValueError(
                "'axes' argument to get_load_settlement_plot() must be a `pyplot.axes.Axes` object or None."
            )
    else:
        kwargs_subplot = {
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        _, axes = plt.subplots(
            1,
            1,
            **kwargs_subplot,
        )

        if not isinstance(axes, Axes):
            raise ValueError(
                "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
            )

    mob_ratio, sb_rb_deq, sb_rs = get_load_settlement_axes_data(settlement_curve)

    # Conversion from sb/Deq [%] to sb [mm]
    # sb_rb = sb/Deq [%] * Deq [m] / 100 [%] * 1000 (to mm)
    sb_rb = sb_rb_deq * 10 * d_eq
    sb_chart = np.append(np.flip(sb_rs), sb_rb)

    # forces
    rb_a = mob_ratio * r_b_k
    rs_a = -mob_ratio * r_s_k
    r_chart = np.append(np.flip(rs_a), rb_a)

    axes.set_ylabel("$s_b$ [mm]")
    axes.set_xlabel(
        "<-- $R_{s,k}$ [kN]                                        $R_{b,k}$ [kN] -->"
    )
    axes.plot(r_chart, sb_chart)
    axes.grid()
    if s_b <= sb_max:
        axes.annotate(
            "",
            xy=(max(-1 * rs_mob_ratio * r_s_k, -r_s_k), s_b),
            xytext=(min(rb_mob_ratio * r_b_k, r_b_k), s_b),
            arrowprops={"arrowstyle": "<|-|>"},
        )
        axes.annotate(
            f"$F_{{c,k,tot}}$: $F_{{c,k}} + F_{{s,k}}$\n$F_{{c,k}}$:{f_c_k: 3.2f} kN\n$F_{{s,k}}$:{f_nk_k: 3.2f} kN",
            xy=(0, s_b),
            xytext=(-60, -60),
            textcoords="offset pixels",
        )
        axes.annotate(
            f"$s_b$ = {s_b:3.2f} mm",
            xy=(min(rb_mob_ratio * r_b_k, r_b_k), s_b),
        )
    else:
        axes.annotate(
            f"$s_b$ = {s_b:3.2f} mm",
            xy=(-25, 0.9 * sb_max),
        )
    axes.axis((-r_s_k, r_b_k, sb_max, 0))

    axes.set_xticks(axes.get_xticks().tolist())  # to avoid FixedLocator warning
    axes.set_xticklabels([str(abs(x)) for x in axes.get_xticks()])

    return axes



from __future__ import annotations

from typing import Dict

MATERIALS = {
    "concrete": {"name": "concrete", "elastic_modulus": 20000, "color": "#525252"},
    "steel": {"name": "steel", "elastic_modulus": 195000, "color": "#E2E2E2"},
    "wood": {"name": "wood", "elastic_modulus": 3600, "color": "#BD7205"},
    "grout": {
        "name": "grout",
        "elastic_modulus": 15000,
        "yield_stress": 1.5,
        "color": "#8A8A8A",
    },
    "grout_extorted": {
        "name": "grout_extorted",
        "elastic_modulus": 20000,
        "yield_stress": 2,
        "color": "#8A8A8A",
    },
}


class Color:
    """The Color class represents an RGB color."""

    def __init__(self, red: int, green: int, blue: int):
        """
        Represents an RGB color.

        Parameters:
        -----------
        red : int
            The red component of the RGB color.
        green : int
            The green component of the RGB color.
        blue : int
            The blue component of the RGB color.
        """
        self.red = red
        self.green = green
        self.blue = blue

    @classmethod
    def from_hex(cls, hex: str) -> Color:
        """
        Instantiate a Color object from a hexadecimal color string.

        Parameters:
        -----------
        hex : str
            The hexadecimal representation of the RGB color.
            example: "#ff0000" for red.

        Returns:
        --------
        Color
            A Color object.
        """
        hex = hex.lstrip("#")
        return cls(
            red=int(hex[0:2], 16),
            green=int(hex[2:4], 16),
            blue=int(hex[4:6], 16),
        )

    @property
    def hex(self) -> str:
        """The hexadecimal representation of the RGB color. e.g. "#ff0000" for red."""
        return "#{:02x}{:02x}{:02x}".format(self.red, self.green, self.blue)

    def serialize_payload(self) -> Dict[str, int]:
        """Serialize the RGB color to a dictionary payload for the API."""
        return {"r": self.red, "g": self.green, "b": self.blue}


class PileMaterial:
    """The PileMaterial class represents a material that can be used in a pile geometry component."""

    def __init__(
        self,
        name: str,
        elastic_modulus: float,
        yield_stress: float | None = None,
        color: Color | str | Dict[str, int] | None = None,
    ):
        """
        Represents a material that can be used in a pile geometry component.

        Parameters:
        -----------
        name : str
            The name of the material.
        elastic_modulus : float
            The elastic modulus [MPa] of the material.
        yield_stress : float, optional
            The yield stress [MPa] of the material, by default None.
        color : Color or str or dict, optional
            The color of the material, by default None.
        """
        self._name = name
        self._elastic_modulus = elastic_modulus
        self._yield_stress = yield_stress
        if isinstance(color, str):
            color = Color.from_hex(hex=color)
        elif isinstance(color, dict):
            color = Color(
                red=color["r"],
                green=color["g"],
                blue=color["b"],
            )
        self._color = color

    @classmethod
    def from_api_response(cls, material: dict) -> PileMaterial:
        """
        Instantiates a PileMaterial from a material object in the API response payload.

        Args:
            material: A dictionary that is retrieved from the API response payload at "/pile_properties/geometry/materials/[i]".
        """
        if isinstance(material["color"], str):
            color = Color.from_hex(hex=material["color"])
        else:
            color = Color(
                red=material["color"]["r"],
                green=material["color"]["g"],
                blue=material["color"]["b"],
            )

        return cls(
            name=material["name"],
            elastic_modulus=material["elastic_modulus"],
            yield_stress=material.get("yield_stress"),
            color=color,
        )

    @property
    def name(self) -> str:
        """The name of the material"""
        return self._name

    @property
    def elastic_modulus(self) -> float:
        """The elastic modulus [MPa] of the material"""
        return self._elastic_modulus

    @property
    def yield_stress(self) -> float | None:
        """The yield stress [MPa] of the material"""
        return self._yield_stress

    @property
    def color(self) -> Color | Color | None:
        """The color of the material"""
        return self._color

    def serialize_payload(self) -> Dict[str, str | float | Dict[str, int]]:
        """
        Serialize the material to a dictionary payload for the API.

        Returns:
            A dictionary payload containing the material's name, elastic modulus, and color.
        """
        payload: Dict[str, str | float | Dict[str, int]] = {
            "name": self.name,
            "elastic_modulus": self.elastic_modulus,
        }

        if self.color is not None:
            payload["color"] = self.color.serialize_payload()

        return payload



from __future__ import annotations  # noqa: F404

from typing import Any

from IPython.display import DisplayHandle, display
from ipywidgets import widgets

from pypilecore.results.cases_multi_cpt_results import CasesMultiCPTBearingResults
from pypilecore.results.result_definitions import CPTResultDefinitions
from pypilecore.viewers.interactive_figures import FigureCPTResultsVersusPtls


class ViewerCptResults:
    """
    Viewer for the CPT results of the bearing capacity calculations.

    It offers the following layout:
        - Dropdown widgets:
            - Case: to select the case to show.
            - Result: to select the result to show.
        - Figure CPT results vs. pile tip level:
            - X axis: result values
            - Y axis: pile tip level w.r.t. NAP
            - Each trace represents a different CPT.
    """

    def __init__(self, cases_multi_results: CasesMultiCPTBearingResults) -> None:
        """Initialize the viewer.

        Parameters
        ----------
        cases_multi_results : CasesMultiCPTBearingResults
            The results of the bearing capacity calculations.

        Raises
        ------
        TypeError
            If 'cases_multi_results' are not of type 'CasesMultiCPTBearingResults'.
        """

        # Initialize figure CPT resuls vs. pile tip level
        self._figure_plts = FigureCPTResultsVersusPtls(
            cases_multi_results=cases_multi_results
        )

        # Set up control widgets
        self._case_dropdown = widgets.Dropdown(
            description="Case:",
            value=self._figure_plts.cases[0],
            options=self._figure_plts.cases,
        )
        self._result_dropdown = widgets.Dropdown(
            description="Result:",
            value=CPTResultDefinitions.R_c_d_net.name,
            options=CPTResultDefinitions.natsorted_names(),
        )

        # Update plot for initial selection of control widgets
        self._update_case_and_result(None)

        # Set up callbacks
        self._case_dropdown.observe(self._update_case_and_result, "value")
        self._result_dropdown.observe(self._update_case_and_result, "value")

        # Set up layout
        self._control_widgets = widgets.HBox(
            [
                self._case_dropdown,
                self._result_dropdown,
            ]
        )
        self._layout = widgets.VBox(
            [self._control_widgets, self._figure_plts.figure]
        )  # , width=800)

    def _update_case_and_result(self, change: Any) -> None:
        """Private method to update the figure when the case or result name are changed in the control widgets."""

        # Update the figure
        self._figure_plts.show_case_and_result(
            case_name=self._case_dropdown.value,
            result_name=self._result_dropdown.value,
        )

    def display(self) -> DisplayHandle | None:
        """Display the figure."""
        return display(self._layout)



from __future__ import annotations

import math
from typing import Any, Literal, Tuple

import matplotlib.patches as patches
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from numpy.typing import NDArray

from pypilecore.common.piles.geometry.components.common import (
    PrimaryPileComponentDimension,
    _BasePileGeometryComponent,
    get_area_vs_depth,
    get_circum_vs_depth,
    get_component_bounds_nap,
    instantiate_axes,
)


def calculate_equiv_tip_diameter(dim1: float, dim2: float | None = None) -> float:
    """
    Calculate the equivalent tip diameter of a rectangular pile.

    Parameters
    ----------
    dim1, dim2 : float
        The dimension of the rectangular pile (in any unit).

    Returns
    -------
    float
        The equivalent tip diameter of the rectangular pile (in the same unit as `dim1` and `dim2`).

    Notes
    -----
    The equivalent tip diameter is calculated based on the dimensions of a rectangular pile. If the maximum dimension
    (`b_max`) is greater than 1.5 times the minimum dimension (`a_min`), the equivalent tip diameter is equal to the
    minimum dimension (`a_min`). Otherwise, it is calculated using the formula: 1.13 * `a_min` * sqrt(`b_max` / `a_min`).

    Examples
    --------
    >>> calculate_equiv_tip_diameter(2.0, 3.0)
    2.0

    >>> calculate_equiv_tip_diameter(2.0, 4.0)
    2.26
    """
    if dim2 is None:
        dim2 = dim1

    b_max = max(dim1, dim2)
    a_min = min(dim1, dim2)

    if b_max > (1.5 * a_min):
        return a_min
    return float(1.13 * a_min * math.sqrt(b_max / (a_min + 1e-12)))


class RectPileGeometryComponent(_BasePileGeometryComponent):
    """The RectPileGeometryComponent class represents a rectangular pile-geometry component."""

    def __init__(
        self,
        secondary_dimension: float,
        primary_dimension: PrimaryPileComponentDimension,
        tertiary_dimension: float | None = None,
        inner_component: _BasePileGeometryComponent | None = None,
        material: str | None = None,
    ):
        """
        Represents a rectangular pile-geometry component.

        Parameters:
        -----------
        secondary_dimension : float
            The secondary dimension [m] of the pile-geometry component, which is the largest cross-sectional dimension.
        primary_dimension : PrimaryPileComponentDimension
            The primary dimension [m] of the pile-geometry component, which is measured along the primary axis of the pile.
        tertiary_dimension : float, optional
            The tertiary dimension [m] of the pile-geometry component, which is the smallest cross-sectional dimension.
        inner_component : RoundPileGeometryComponent | RectPileGeometryComponent | None, optional
            The component on the inside of the pile-geometry component, by default None.
        material : str, optional
            The material name of the pile-geometry component, by default None.
        """
        self._secondary_dimension = secondary_dimension
        self._tertiary_dimension = tertiary_dimension
        self._primary_dimension = primary_dimension
        self._inner_component = inner_component
        self._material = material

    @classmethod
    def from_api_response(
        cls,
        component: dict,
        inner_component: _BasePileGeometryComponent | None = None,
    ) -> RectPileGeometryComponent:
        """
        Instantiates a RectPileGeometryComponent from a component object in the API
        response payload.

        Parameters:
        -----------
        component: dict
            A dictionary that is retrieved from the API response payload at "/pile_properties/geometry/components/[i]".
            The component dictionary should have the following schema:

            {
                "secondary_dimension": float,
                "tertiary_dimension": float,
                "primary_dimension": {
                    "length": float,
                    "width": float
                },
                "material": str
            }

            - secondary_dimension (float): The secondary dimension of the rectangular pile component.
              This dimension represents the dimension perpendicular to the primary dimension.
            - tertiary_dimension (float): The tertiary dimension of the rectangular pile component.
              This dimension represents the dimension perpendicular to both the primary and secondary dimensions.
            - primary_dimension (dict): The primary dimension of the rectangular pile component.
                - length (float): The length of the rectangular pile component.
                  This dimension represents the length of the pile along its primary axis.
                - width (float): The width of the rectangular pile component.
                  This dimension represents the width of the pile along its secondary axis.
            - material (str): The material of the rectangular pile component.

        inner_component: RectPileGeometryComponent | RoundPileGeometryComponent | None, optional
            The component on the inside of the pile-geometry component, by default None.

        Returns:
        --------
        RectPileGeometryComponent
            A rectangular pile-geometry component.

        Example:
        --------
        >>> component = {
        ...     "secondary_dimension": 10,
        ...     "tertiary_dimension": 5,
        ...     "primary_dimension": {
        ...         "length": 20,
        ...         "width": 15
        ...     },
        ...     "material": "concrete"
        ... }
        >>> inner_component = None
        >>> result = RectPileGeometryComponent.from_api_response(component, inner_component)
        """
        return cls(
            secondary_dimension=component["secondary_dimension"],
            tertiary_dimension=component["tertiary_dimension"],
            primary_dimension=PrimaryPileComponentDimension.from_api_response(
                component["primary_dimension"]
            ),
            inner_component=inner_component,
            material=component["material"],
        )

    @property
    def inner_component(
        self,
    ) -> _BasePileGeometryComponent | None:
        """The component on the inside of the pile-geometry component"""
        return self._inner_component

    @property
    def outer_shape(self) -> Literal["rectangle"]:
        """The outer shape of the pile-geometry component"""
        return "rectangle"

    @property
    def material(self) -> str | None:
        """The material name of the pile-geometry component"""
        return self._material

    @property
    def primary_dimension(self) -> PrimaryPileComponentDimension:
        """
        The primary dimension [m] of the pile-geometry component, which is measured along the primary axis of the pile.
        """
        return self._primary_dimension

    @property
    def secondary_dimension(self) -> float:
        """
        The secondary dimension [m] of the pile-geometry component, which is the largest cross-sectional dimension.
        """
        return self._secondary_dimension

    @property
    def tertiary_dimension(self) -> float:
        """
        The tertiary dimension [m] of the pile-geometry component, which is the smallest cross-sectional dimension.
        """
        if self._tertiary_dimension is not None:
            return self._tertiary_dimension
        return self.secondary_dimension

    @property
    def cross_section_bounds(self) -> Tuple[float, float, float, float]:
        """Alias of the diameter [m] of the pile-geometry component"""
        return (
            -self.secondary_dimension / 2,
            self.secondary_dimension / 2,
            -self.tertiary_dimension / 2,
            self.tertiary_dimension / 2,
        )

    @property
    def circumference(self) -> float:
        """The outer-circumference [m] of the pile-geometry component"""
        return 2 * (self.secondary_dimension + self.tertiary_dimension)

    @property
    def equiv_tip_diameter(self) -> float:
        """
        Equivalent outer-diameter [m] of the component at the tip-level.
        According to NEN-9997-1+C2_2017 paragraphs 1.5.2.106a and 7.6.2.3.(10)(e).

        Specifically: returns self.tertiary_dimension
        if self.primary_dimension > (1,5 * self.tertiary_dimension)
        """
        return calculate_equiv_tip_diameter(
            self.tertiary_dimension, self.secondary_dimension
        )

    @property
    def area_full(self) -> float:
        """The full outer-area [m²] of the pile-geometry component, including any potential inner-components"""
        return self.secondary_dimension * self.tertiary_dimension

    def serialize_payload(self) -> dict:
        """
        Serialize the rectangular pile-geometry component to a dictionary payload for the API.

        Returns:
            A dictionary payload containing the outer shape, secondary dimension, tertiary dimension (if set), material, and primary dimension (if set).
        """
        payload = {
            "outer_shape": self.outer_shape,
            "primary_dimension": self.primary_dimension.serialize_payload(),
            "secondary_dimension": self.secondary_dimension,
            "material": self.material,
        }
        if self.tertiary_dimension is not None:
            payload["tertiary_dimension"] = self.tertiary_dimension

        return payload

    def get_component_bounds_nap(
        self,
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> Tuple[float, float]:
        """
        Returns component head and tip level in NAP.

        Parameters
        ----------
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        tuple
            Tuple with component head and tip level in [m] w.r.t. NAP.
        """
        return get_component_bounds_nap(
            pile_tip_level_nap=pile_tip_level_nap,
            pile_head_level_nap=pile_head_level_nap,
            component_primary_length=self.primary_dimension.length,
        )

    def get_circum_vs_depth(
        self,
        depth_nap: NDArray[np.floating],
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> NDArray[np.floating]:
        """
        Returns component circumferences at requested depths.

        Parameters
        ----------
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with component circumferences at the depths in the depth parameter.
        """
        return get_circum_vs_depth(
            depth_nap=depth_nap,
            pile_tip_level_nap=pile_tip_level_nap,
            pile_head_level_nap=pile_head_level_nap,
            length=self.primary_dimension.length,
            circumference=self.circumference,
        )

    def get_inner_area_vs_depth(
        self,
        depth_nap: NDArray[np.floating],
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> NDArray[np.floating]:
        """
        Returns inner component areas at requested depths.

        Parameters
        ----------
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with inner component areas at the depths in the depth parameter.
        """
        if self.inner_component is None:
            return np.zeros_like(depth_nap)

        return self.inner_component.get_area_vs_depth(
            depth_nap=depth_nap,
            pile_tip_level_nap=pile_tip_level_nap,
            pile_head_level_nap=pile_head_level_nap,
        )

    def get_area_vs_depth(
        self,
        depth_nap: NDArray[np.floating],
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> NDArray[np.floating]:
        """
        Returns component areas at requested depths.

        Parameters
        ----------
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with component areas at the depths in the depth parameter.
        """

        (
            component_head_level_nap,
            component_tip_level_nap,
        ) = self.get_component_bounds_nap(pile_tip_level_nap, pile_head_level_nap)

        inner_area = self.get_inner_area_vs_depth(
            depth_nap=depth_nap,
            pile_tip_level_nap=pile_tip_level_nap,
            pile_head_level_nap=pile_head_level_nap,
        )

        return get_area_vs_depth(
            depth_nap=depth_nap,
            area_full=self.area_full,
            component_head_level_nap=component_head_level_nap,
            component_tip_level_nap=component_tip_level_nap,
            inner_area=inner_area,
        )

    def plot_cross_section_exterior(
        self,
        figsize: Tuple[float, float] = (6.0, 6.0),
        facecolor: Tuple[float, float, float] | str | None = None,
        axes: Axes | None = None,
        axis_arg: bool | str | Tuple[float, float, float, float] | None = "auto",
        show: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the cross-section of the pile at a specified depth.

        Parameters
        ----------
        figsize : tuple, optional
            The figure size (width, height) in inches, by default (6.0, 6.0).
        facecolor : tuple or str, optional
            The face color of the pile cross-section, by default None.
        axes : Axes
            The axes object to plot the cross-section on.
        axis_arg : bool or str or tuple, optional
            The axis argument to pass to the `axes.axis()` function, by default "auto".
        show : bool, optional
            Whether to display the plot, by default True.
        **kwargs
            Additional keyword arguments to pass to the `plt.subplots()` function.
        """
        axes = instantiate_axes(
            figsize=figsize,
            axes=axes,
            **kwargs,
        )

        x_offset = -self.secondary_dimension / 2
        y_offset = -self.tertiary_dimension / 2

        axes.add_patch(
            patches.Rectangle(
                (x_offset, y_offset),
                self.secondary_dimension,
                self.tertiary_dimension,
                facecolor=facecolor,
                edgecolor="black",
            )
        )
        if axis_arg:
            axes.axis(axis_arg)
        if show:
            plt.show()
        return axes

    def plot_side_view(
        self,
        bottom_boundary_nap: float | Literal["pile_tip"] = "pile_tip",
        top_boundary_nap: float | Literal["pile_head"] = "pile_head",
        pile_tip_level_nap: float | int = -10,
        pile_head_level_nap: float | int = 0,
        figsize: Tuple[float, float] = (6.0, 6.0),
        facecolor: Tuple[float, float, float] | str | None = None,
        axes: Axes | None = None,
        axis_arg: bool | str | Tuple[float, float, float, float] | None = "scaled",
        show: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the side view of the component at a specified depth.

        Parameters
        ----------
        bottom_boundary_nap : float or str, optional
            The bottom boundary level of the plot, in m w.r.t. NAP. Default = "pile_tip".
        top_boundary_nap : float or str, optional
            The top boundary level of the plot, in m w.r.t. NAP. Default = "pile_head".
        pile_tip_level_nap : float, optional
            The pile tip level in m w.r.t. NAP. Default = -10.
        pile_head_level_nap : float, optional
            The pile head level in m w.r.t. NAP. Default = 0.
        figsize : tuple, optional
            The figure size (width, height) in inches, by default (6.0, 6.0).
        facecolor : tuple or str, optional
            The face color of the pile cross-section, by default None.
        axes : Axes
            The axes object to plot the cross-section on.
        axis_arg : bool or str or tuple, optional
            The axis argument to pass to the `axes.axis()` function, by default "auto".
        show : bool, optional
            Whether to display the plot, by default True.
        **kwargs
            Additional keyword arguments to pass to the `plt

        Returns
        -------
        Axes
            The axes object to plot the cross-section on.
        """
        axes = instantiate_axes(
            figsize=figsize,
            axes=axes,
            **kwargs,
        )

        if top_boundary_nap == "pile_head":
            top_boundary_nap = pile_head_level_nap

        if bottom_boundary_nap == "pile_tip":
            bottom_boundary_nap = pile_tip_level_nap

        (
            component_head_level_nap,
            component_tip_level_nap,
        ) = self.get_component_bounds_nap(pile_tip_level_nap, pile_head_level_nap)

        if (
            top_boundary_nap > component_tip_level_nap
            and bottom_boundary_nap < component_head_level_nap
        ):
            z_offset = component_head_level_nap
            height = (
                max(component_tip_level_nap, bottom_boundary_nap)
                - component_head_level_nap
            )

            axes.add_patch(
                patches.Rectangle(
                    (self.cross_section_bounds[0], z_offset),
                    self.secondary_dimension,
                    height,
                    facecolor=facecolor,
                )
            )

        if axis_arg:
            axes.axis(axis_arg)
        if show:
            plt.show()
        return axes



from __future__ import annotations

import logging
from copy import deepcopy
from typing import Any, Dict, List, Literal

import numpy as np
from shapely.geometry import Polygon, mapping

from pypilecore.results import SingleCPTBearingResults

_dft_optimize_result_by = [
    "minimum_pile_level",
    "number_of_cpts",
    "number_of_consecutive_pile_levels",
]


def create_grouper_payload(
    cpt_results_dict: Dict[str, SingleCPTBearingResults],
    pile_load_uls: float,
    building_polygon: Polygon | None = None,
    cpt_grid_rotation: float = 0.0,
    gamma_bottom: float = 1.2,
    gamma_shaft: float = 1.2,
    include_centre_to_centre_check: bool = False,
    stiff_construction: bool = False,
    optimize_result_by: List[
        Literal[
            "minimum_pile_level",
            "number_of_cpts",
            "number_of_consecutive_pile_levels",
            "centre_to_centre_check",
        ]
    ]
    | None = _dft_optimize_result_by,  # type: ignore
    resolution: float = 0.5,
    overrule_nan: float = 0.0,
    skip_nan: bool = False,
) -> dict:
    """
    Creates a dictionary with the payload content for the PileCore endpoint
    "/grouper/group_cpts"


    Note
    ------
    The grouper uses pile bearing capacity results calculated by PileCore or other software to
    form  subgroups of the total group of CPT’s belonging to this project.
    Valid subgroups have three characteristics:

        - a maximum variation coefficient of 12% at one or more pile-tip levels. (Variation check
          NEN9997-1 A.3.3.3)
        - a minimum design pile bearing capacity based on the given pile load ULS at one or more
          pile-tip levels. (Bearing check)
        - is spatially coherent, which means there are no other CPTs in between the members
          of the subgroup. (Spatial check)

    Additionally, centre to centre validation (include_centre_to_centre_check; NEN9997-1 3.2.3) can be
    added to the cluster method. This check adds restrictions to the maximum allowable R;c;cal outliers
    and makes sure that the suitable data density requirements for the subgroup are met, by checking the
    centre-to-centre (CTC) distance of the cpts in a regular square grid:

        - A maximum CTC distance of 25 m, if no R;c;cal outliers greater than 30% off the average;
        - A maximum CTC distance of 20 m, if no R;c;cal outliers greater than 40% off the average;
        - A maximum CTC distance of 15 m, if no R;c;cal outliers greater than 50% off the average;
        - All subgroups with R;c;cal outliers greater than 50% of the average are considered invalid.

    The CTC check is performed by drawing a square around each CPT with the CTC dimensions according to the
    outlier criterion and verifying that there is no empty space between squares. The rotation of the squares
    can be provided with the `cpt_grid_rotation` argument. Note that this rotation is assigned to all CPT's and
    therefore should represent the main orientation of the building.

    Parameters
    ----------
    cpt_results_dict:
        Dictionary with key as CPT name and value a SingleCPTBearingResults class.
        Should contain at least 2 entries.
    pile_load_uls
        ULS load in kN. Used to determine if a grouping configuration is valid.
    stiff_construction
        Default is False
        Attribute use to get the xi3 and xi4 value. True if it is a stiff construction
    optimize_result_by
        Default is "minimum_pile_level", "number_of_cpts", "number_of_consecutive_pile_levels"
        Attribute that states how to sort the result and find groups.
        Based on the filter method, a selection of valid subgroups are included in the report. The following
        filters are available:

            - Number_of_cpts: the grouper adds filters to make the group as big as possible to try and get a
              uniform pile tip level for most CPT’s.
            - Number_of_consecutive_pile_levels; the grouper adds filters to get groups that contain consecutive
              pile tip levels to ensure a consistent soil layer is used.
            - Minimum_pile_level; the grouper adds filters to return groups that optimize pile length to try and
              optimize or reduce material use.
            - Centre_to_centre_check; the grouper adds filters to favour groups that are valid according to the
              centre to centre rules of the NEN9997-1 3.2.3.

    gamma_shaft
        Default is 1.2
        Safety factor shaft design bearing capacity
    gamma_bottom
        Default is 1.2
        Safety factor bottom design bearing capacity
    include_centre_to_centre_check:
        Default is False
        Flag that indicates if the cluster algorithm performs a centre to centre validation of the CPT’s of the
        generated subgroups or not according to NEN9997-1 3.2.3. If the group doesn’t comply to the check, the
        result of the group is deleted from the result.
    building_polygon:
        Default is None
        Polygon of the contour of the building. If None the building Polygon in generated based on the
        convex_hull of the CPT points.
    cpt_grid_rotation:
        Default is 0.0.
        Rotation of the squares used in the centre to centre validation [degrees]
    resolution:
        Default is 0.5
        The resolution of clusters algorithm. If resolution is 1 the cluster boundary conditions can be met
        (number clusters is number CPTs). Depending on the number of CPTs this can take some time.
    overrule_nan:
        Default is 0.0
        The default behavior is to replace NaN with zero, for one of the following
        attributes ["R_b_cal", "F_nk_d", "R_s_cal"].
    skip_nan:
        Default is False
        If True the CPTs are skipped that have NaN values in one of the following
        attributes ["R_b_cal", "F_nk_d", "R_s_cal"], this means that they are not used in the grouper method.

    Raises
    ------
    ValueError:
        - if NaN values are present in negative friction, bottom or shaft bearing_capacity

        - if x or y coordinate is None

        - if pile tip levels don't macht for all SingleCPTBearingResults

        - if less than 2 valid CPTs are provided


    Returns
    -------
    payload:
        Dictionary with the payload content for the PileCore endpoint
        "/grouper/group_cpts"
    """
    # create default payload object
    payload: Dict[str, Any] = {
        "cpt_grid_rotation": cpt_grid_rotation,
        "gamma_bottom": gamma_bottom,
        "gamma_shaft": gamma_shaft,
        "include_centre_to_centre_check": include_centre_to_centre_check,
        "pile_load_uls": pile_load_uls,
        "stiff_construction": stiff_construction,
        "resolution": resolution,
        "optimize_result_by": optimize_result_by,
    }

    # set source building polygon in payload
    if building_polygon is not None:
        payload["building_polygon"] = mapping(building_polygon)

    # set bearing capacity in payload
    cpt_objects = []
    pile_tip_level_object = {}
    for name, cpt_result in cpt_results_dict.items():
        has_nan = False
        # check if coordinate are set
        if cpt_result.soil_properties.x is None or cpt_result.soil_properties.y is None:
            raise ValueError(
                f" CPT {name} does not have a x-coordinate or y-coordinate"
            )

        for item in ["R_b_cal", "F_nk_d", "R_s_cal"]:
            if np.isnan(cpt_result.table.__getattribute__(item)).any():
                if skip_nan:
                    has_nan = True
                    logging.warning(
                        f"CPT {name} has NaN values are present in column {item}. "
                        f"Not included in grouper payload."
                    )

                    break
                else:
                    logging.warning(
                        f"CPT {name} has NaN values are present in column {item}. "
                        f"Replace NaN with {overrule_nan}."
                    )
        # skip CPT that are not valid.
        if has_nan:
            continue

        # map pile tip levels to object
        pile_tip_level_object[name] = cpt_result.table.pile_tip_level_nap.tolist()

        # add bearing capacity result to object
        cpt_objects.append(
            {
                "bottom_bearing_capacity": np.nan_to_num(
                    cpt_result.table.R_b_cal, nan=overrule_nan
                ).tolist(),
                "negative_friction": np.nan_to_num(
                    cpt_result.table.F_nk_d, nan=overrule_nan
                ).tolist(),
                "shaft_bearing_capacity": np.nan_to_num(
                    cpt_result.table.R_s_cal, nan=overrule_nan
                ).tolist(),
                "name": name,
                "coordinates": {
                    "x": cpt_result.soil_properties.x,
                    "y": cpt_result.soil_properties.y,
                },
            }
        )

    if not len(cpt_objects) >= 2:
        raise ValueError(
            "The PileCore grouper requires at least 2 CPTs with valid bearing capacity."
        )

    payload["cpt_objects"] = cpt_objects

    # validate pile tip levels
    raw_lengths = [
        frozenset(np.round(values, 1)) for values in pile_tip_level_object.values()
    ]
    if len(list(set(raw_lengths))) > 1:
        msg = "The PileCore grouper requires all CPTs to have a valid bearing capacity for all pile tip levels. \n"
        for name, pile_tip_level in pile_tip_level_object.items():
            msg += (
                f"Pile tip levels are not similar for CPT {name} with length {len(pile_tip_level)}, "
                f"upper boundary: {max(pile_tip_level)}, lower boundary: {min(pile_tip_level)}. \n"
            )
        raise ValueError(msg)
    payload["pile_tip_level"] = sorted(list(raw_lengths[0]), reverse=True)

    return payload


def create_grouper_report_payload(
    grouper_payload: dict,
    grouper_response: dict,
    project_name: str,
    project_id: str,
    author: str,
) -> dict:
    """
    Creates a dictionary with the payload content for the PileCore endpoint
    "grouper/generate_grouper_report"

    Parameters
    ----------
    grouper_payload:
       The resulting payload of a call to `create_grouper_payload()`
    grouper_response:
       The resulting response of a call to `get_groups_api_result()`
    project_name:
        The name of the project.
    project_id:
        The identifier (code) of the project.
    author:
        The author of the report.

    Returns
    -------
    report_payload:
        Dictionary with the payload content for the PileCore endpoint
        "grouper/generate_grouper_report"
    """
    report_payload = deepcopy(grouper_payload)
    report_payload.update(
        dict(
            sub_groups=grouper_response["sub_groups"],
            author=author,
            project_number=project_id,
            project_name=project_name,
        ),
    )
    # remove not used attributes
    _ = report_payload.pop("pile_tip_level")
    _ = report_payload.pop("cpt_objects")
    return report_payload



from __future__ import annotations  # noqa: F404

from dataclasses import dataclass
from enum import Enum
from typing import List

from natsort import natsorted


@dataclass
class ResultDefinition:
    """
    Dataclass containing the name, units and html representation of a result.
    """

    name: str
    """The name of the result."""
    unit: str
    """The unit of the result."""
    html: str
    """The html representation of the result"""

    def __post_init__(self) -> None:
        """
        Method to validate the input types of the dataclass.

        Raises
        ------
        ValueError
            If the input types are not correct.
        """
        if not isinstance(self.name, str):
            raise TypeError(
                f"Expected type 'str' for 'name', but got {type(self.name)}"
            )
        if not isinstance(self.unit, str):
            raise TypeError(
                f"Expected type 'str' for 'unit', but got {type(self.unit)}"
            )

        if not isinstance(self.html, str):
            raise TypeError(
                f"Expected type 'str' for 'html', but got {type(self.html)}"
            )


class CPTResultDefinitions(Enum):
    """
    Enumeration of available CPT result definitions.
    """

    F_nk_cal = ResultDefinition(name="F_nk_cal", unit="kN", html="F<sub>nk;cal</sub>")
    F_nk_k = ResultDefinition(name="F_nk_k", unit="kN", html="F<sub>nk;k</sub>")
    F_nk_d = ResultDefinition(name="F_nk_d", unit="kN", html="F<sub>nk;d</sub>")
    R_b_cal = ResultDefinition(name="R_b_cal", unit="kN", html="R<sub>b;cal</sub>")
    R_b_k = ResultDefinition(name="R_b_k", unit="kN", html="R<sub>b;k</sub>")
    R_b_d = ResultDefinition(name="R_b_d", unit="kN", html="R<sub>b;d</sub>")
    R_s_cal = ResultDefinition(name="R_s_cal", unit="kN", html="R<sub>s;cal</sub>")
    R_s_k = ResultDefinition(name="R_s_k", unit="kN", html="R<sub>s;k</sub>")
    R_s_d = ResultDefinition(name="R_s_d", unit="kN", html="R<sub>s;d</sub>")
    R_c_cal = ResultDefinition(name="R_c_cal", unit="kN", html="R<sub>c;cal</sub>")
    R_c_k = ResultDefinition(name="R_c_k", unit="kN", html="R<sub>c;k</sub>")
    R_c_d = ResultDefinition(name="R_c_d", unit="kN", html="R<sub>c;d</sub>")
    R_c_d_net = ResultDefinition(
        name="R_c_d_net",
        unit="kN",
        html="R<sub>c;d;net</sub>",
    )
    F_c_k = ResultDefinition(name="F_c_k", unit="kN", html="F<sub>c;k</sub>")
    F_c_k_tot = ResultDefinition(
        name="F_c_k_tot",
        unit="kN",
        html="F<sub>c;k;tot</sub>",
    )
    negative_friction_range_nap_top = ResultDefinition(
        name="negative_friction_range_nap_top",
        unit="m NAP",
        html="Top of negative friction",
    )
    negative_friction_range_nap_btm = ResultDefinition(
        name="negative_friction_range_nap_btm",
        unit="m NAP",
        html="Bottom of negative friction",
    )
    positive_friction_range_nap_top = ResultDefinition(
        name="positive_friction_range_nap_top",
        unit="m NAP",
        html="Top of positive friction",
    )
    positive_friction_range_nap_btm = ResultDefinition(
        name="positive_friction_range_nap_btm",
        unit="m NAP",
        html="Bottom of positive friction",
    )
    q_b_max = ResultDefinition(name="q_b_max", unit="MPa", html="q<sub>b;max</sub>")
    q_s_max_mean = ResultDefinition(
        name="q_s_max_mean",
        unit="MPa",
        html="q<sub>s;max</sub>",
    )
    qc1 = ResultDefinition(name="qc1", unit="MPa", html="q<sub>c1</sub>")
    qc2 = ResultDefinition(name="qc2", unit="MPa", html="q<sub>c2</sub>")
    qc3 = ResultDefinition(name="qc3", unit="MPa", html="q<sub>c3</sub>")
    s_b = ResultDefinition(name="s_b", unit="mm", html="s<sub>b</sub>")
    s_el = ResultDefinition(name="s_el", unit="mm", html="s<sub>el</sub>")
    k_v_b = ResultDefinition(name="k_v_b", unit="MN/m", html="k<sub>v;b</sub>")
    k_v_1 = ResultDefinition(name="k_v_1", unit="MN/m", html="k<sub>v;1</sub>")

    @classmethod
    def get(cls, name: str) -> CPTResultDefinitions:
        """Returns the result definition the given name."""
        try:
            return cls[name]
        except KeyError:
            raise ValueError(
                f"Result with name '{name}' not found in 'CPTResultDefinitions'."
            )

    @classmethod
    def natsorted_names(cls) -> List[str]:
        """Returns the names of the enum in natsorted order."""
        return natsorted([r.name for r in cls])


class CPTGroupResultDefinitions(Enum):
    R_s_k = ResultDefinition(name="R_s_k", unit="kN", html="R<sub>s;k</sub>")
    R_b_k = ResultDefinition(name="R_b_k", unit="kN", html="R<sub>b;k</sub>")
    R_c_k = ResultDefinition(name="R_c_k", unit="kN", html="R<sub>c;k</sub>")
    R_s_d = ResultDefinition(name="R_s_d", unit="kN", html="R<sub>s;d</sub>")
    R_b_d = ResultDefinition(name="R_b_d", unit="kN", html="R<sub>b;d</sub>")
    R_c_d = ResultDefinition(name="R_c_d", unit="kN", html="R<sub>c;d</sub>")
    F_nk_cal_mean = ResultDefinition(
        name="F_nk_cal_mean", unit="kN", html="F<sub>nk;cal;mean</sub>"
    )
    F_nk_k = ResultDefinition(name="F_nk_k", unit="kN", html="F<sub>nk;k</sub>")
    F_nk_d = ResultDefinition(name="F_nk_d", unit="kN", html="F<sub>nk;d</sub>")
    R_c_d_net = ResultDefinition(
        name="R_c_d_net",
        unit="kN",
        html="R<sub>c;d;net</sub>",
    )
    F_c_k = ResultDefinition(name="F_c_k", unit="kN", html="F<sub>c;k</sub>")
    F_c_k_tot = ResultDefinition(
        name="F_c_k_tot",
        unit="kN",
        html="F<sub>c;k;tot</sub>",
    )
    s_b = ResultDefinition(name="s_b", unit="mm", html="s<sub>b</sub>")
    s_e = ResultDefinition(name="s_e", unit="mm", html="s<sub>e</sub>")
    s_e_mean = ResultDefinition(name="s_e_mean", unit="mm", html="s<sub>e;mean</sub>")
    R_b_mob_ratio = ResultDefinition(
        name="R_b_mob_ratio", unit="-", html="R<sub>b;mob;ratio</sub>"
    )
    R_s_mob_ratio = ResultDefinition(
        name="R_s_mob_ratio", unit="-", html="R<sub>s;mob;ratio</sub>"
    )
    k_v_b = ResultDefinition(name="k_v_b", unit="MN/m", html="k<sub>v;b</sub>")
    k_v_1 = ResultDefinition(name="k_v_1", unit="MN/m", html="k<sub>v;1</sub>")
    R_c_min = ResultDefinition(name="R_c_min", unit="kN", html="R<sub>c;min</sub>")
    R_c_max = ResultDefinition(name="R_c_max", unit="kN", html="R<sub>c;max</sub>")
    R_c_mean = ResultDefinition(name="R_c_mean", unit="kN", html="R<sub>c;mean</sub>")
    R_c_std = ResultDefinition(name="R_c_std", unit="kN", html="R<sub>c;std</sub>")
    R_s_mean = ResultDefinition(name="R_s_mean", unit="kN", html="R<sub>s;mean</sub>")
    R_b_mean = ResultDefinition(name="R_b_mean", unit="kN", html="R<sub>b;mean</sub>")
    var_coef = ResultDefinition(name="var_coef", unit="%", html="Variation coefficient")
    n_cpts = ResultDefinition(name="n_cpts", unit="-", html="Number of CPTs")
    use_group_average = ResultDefinition(
        name="use_group_average", unit="-", html="Use group average"
    )
    xi_normative = ResultDefinition(name="xi_normative", unit="-", html="Normative ξ")
    xi_value = ResultDefinition(name="xi_value", unit="-", html="ξ<sub>value</sub>")
    cpt_Rc_min = ResultDefinition(
        name="cpt_Rc_min", unit="-", html="CPT with R<sub>c;min</sub>"
    )
    cpt_Rc_max = ResultDefinition(
        name="cpt_Rc_max", unit="-", html="CPT with R<sub>c;max</sub>"
    )
    cpt_normative = ResultDefinition(
        name="cpt_normative", unit="-", html="Normative CPT"
    )

    @classmethod
    def get(cls, name: str) -> CPTGroupResultDefinitions:
        """Returns the result definition given the name."""
        try:
            return cls[name]
        except KeyError:
            raise ValueError(
                f"Result with name '{name}' not found in 'CPTGroupResultDefinitions'."
            )

    @classmethod
    def natsorted_names(cls) -> List[str]:
        """Returns the names of the enum in natsorted order."""
        return natsorted([r.name for r in cls])



from __future__ import annotations

from typing import Dict


class PileType:
    """The PileType class represents the type of a pile."""

    def __init__(
        self,
        standard_pile: Dict[str, str | int] | None = None,
        alpha_s_sand: float | None = None,
        alpha_s_clay: float | None = None,
        alpha_p: float | None = None,
        alpha_t_sand: float | None = None,
        settlement_curve: int | None = None,
        negative_fr_delta_factor: float | None = None,
        adhesion: float | None = None,
        is_low_vibrating: bool | None = None,
        is_auger: bool | None = None,
    ):
        """
        Represents the type of a pile.

        Parameters:
        -----------
        standard_pile : dict, optional
            The standard pile definition of the pile type (if applicable), by default None.
        alpha_s_sand : float, optional
            The alpha_s_sand value of the pile type, by default None.
        alpha_s_clay : float, optional
            The alpha_s_clay value of the pile type, by default None.
        alpha_p : float, optional
            The alpha_p value of the pile type, by default None.
        alpha_t_sand : float, optional
            The alpha_t_sand value of the pile type, by default None.
        settlement_curve : int, optional
            The settlement curve of the pile type, by default None.
        negative_fr_delta_factor : float, optional
            The negative_fr_delta_factor value of the pile type, by default None.
        adhesion : float, optional
            The adhesion value of the pile type, by default None.
        is_low_vibrating : bool, optional
            The is_low_vibrating value of the pile type, by default None.
        is_auger : bool, optional
            The is_auger value of the pile type, by default None.
        """
        self._standard_pile = standard_pile
        self._alpha_s_sand = alpha_s_sand
        self._alpha_s_clay = alpha_s_clay
        self._alpha_p = alpha_p
        self._alpha_t_sand = alpha_t_sand
        self._settlement_curve = settlement_curve
        self._negative_fr_delta_factor = negative_fr_delta_factor
        self._adhesion = adhesion
        self._is_low_vibrating = is_low_vibrating
        self._is_auger = is_auger

    @classmethod
    def from_api_response(cls, pile_type: dict) -> PileType:
        """
        Instantiates a PileType from a pile type object in the API response payload.

        Parameters:
        -----------
        pile_type: dict
            A dictionary that is retrieved from the API response payload at "/pile_properties/pile_type".

        Returns:
        --------
        PileType
            A pile type.
        """
        return cls(
            standard_pile=pile_type.get("standard_pile"),
            alpha_s_sand=pile_type["properties"]["alpha_s_sand"],
            alpha_s_clay=pile_type["properties"]["alpha_s_clay"],
            alpha_p=pile_type["properties"]["alpha_p"],
            alpha_t_sand=pile_type["properties"]["alpha_t_sand"],
            settlement_curve=pile_type["properties"]["settlement_curve"],
            negative_fr_delta_factor=pile_type["properties"][
                "negative_fr_delta_factor"
            ],
            adhesion=pile_type["properties"]["adhesion"],
            is_low_vibrating=pile_type["properties"]["is_low_vibrating"],
            is_auger=pile_type["properties"]["is_auger"],
        )

    @property
    def standard_pile(self) -> Dict[str, str | int] | None:
        """The standard pile definition of the pile type (if applicable)"""
        return self._standard_pile

    @property
    def alpha_s_sand(self) -> float | None:
        """The alpha_s_sand value of the pile type"""
        return self._alpha_s_sand

    @property
    def alpha_s_clay(self) -> float | None:
        """The alpha_s_clay value of the pile type"""
        return self._alpha_s_clay

    @property
    def alpha_p(self) -> float | None:
        """The alpha_p value of the pile type"""
        return self._alpha_p

    @property
    def alpha_t_sand(self) -> float | None:
        """The alpha_t_sand value of the pile type"""
        return self._alpha_t_sand

    @property
    def settlement_curve(self) -> int | None:
        """The settlement curve of the pile type"""
        return self._settlement_curve

    @property
    def negative_fr_delta_factor(self) -> float | None:
        """The negative_fr_delta_factor value of the pile type"""
        return self._negative_fr_delta_factor

    @property
    def adhesion(self) -> float | None:
        """The adhesion value of the pile type"""
        return self._adhesion

    @property
    def is_low_vibrating(self) -> bool | None:
        """The is_low_vibrating value of the pile type"""
        return self._is_low_vibrating

    @property
    def is_auger(self) -> bool | None:
        """The is_auger value of the pile type"""
        return self._is_auger

    def serialize_payload(self) -> Dict[str, str | dict]:
        """
        Serialize the pile type to a dictionary payload for the API.

        Returns:
            A dictionary payload containing the standard pile (if set), alpha_s_sand, alpha_s_clay, alpha_p,
            alpha_t_sand, settlement curve, negative_fr_delta_factor, adhesion, is_low_vibrating, and is_auger.
        """
        payload: Dict[str, str | dict] = {}

        if self.standard_pile is not None:
            payload["standard_pile"] = {
                "main_type": str(self.standard_pile["main_type"]),
                "specification": str(self.standard_pile["specification"]),
            }
            if (
                "installation" in self.standard_pile
                and self.standard_pile["installation"] is not None
            ):
                payload["standard_pile"]["installation"] = str(  # type: ignore
                    self.standard_pile["installation"]
                )

        custom_type_properties: Dict[str, float | bool] = {}

        if self.alpha_s_sand is not None:
            custom_type_properties["alpha_s_sand"] = self.alpha_s_sand

        if self.alpha_s_clay is not None:
            custom_type_properties["alpha_s_clay"] = self.alpha_s_clay

        if self.alpha_p is not None:
            custom_type_properties["alpha_p"] = self.alpha_p

        if self.alpha_t_sand is not None:
            custom_type_properties["alpha_t_sand"] = self.alpha_t_sand

        if self.settlement_curve is not None:
            custom_type_properties["settlement_curve"] = self.settlement_curve

        if self.negative_fr_delta_factor is not None:
            custom_type_properties[
                "negative_fr_delta_factor"
            ] = self.negative_fr_delta_factor

        if self.adhesion is not None:
            custom_type_properties["adhesion"] = self.adhesion

        if self.is_low_vibrating is not None:
            custom_type_properties["is_low_vibrating"] = self.is_low_vibrating

        if self.is_auger is not None:
            custom_type_properties["is_auger"] = self.is_auger

        if len(custom_type_properties.keys()) > 0:
            payload["custom_properties"] = custom_type_properties

        return payload
